/*****************************************************************************
 *
 * FILE:	ping4.c
 * DESCRIPTION:	ping
 * DATE:	Wed, Jun 11 2003
 * UPDATED:	Mon, Mar 17 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: ping4.c,v 1.1 2008/03/18 13:31:39 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	PHDR_LEN	((int)sizeof(struct timeval))
#define	DEFDATALEN	(64 - PHDR_LEN)	/* default data length */
#define	MAXIPLEN	(sizeof(struct ip) + MAX_IPOPTLEN)
#define	MINICMPLEN	ICMP_MINLEN
#define	MAXPAYLOAD	(IP_MAXPACKET - MAXIPLEN - MINICMPLEN)
#define	PING_COUNT	(5)
#define	PING_INTERVAL	(1)	/* sec */
#define	PING_TIMEOUT	(PING_INTERVAL * 1000)	/* msec */

typedef struct {
  void			(*fproc)(char *, ssize_t, struct timeval *);
  void			(*fsend)(void);
  struct sockaddr *	sasend;	/* sockaddr{} for send, from getaddrinfo */
  struct sockaddr *	sarecv;	/* sockaddr{} fro receiving */
  socklen_t		salen;	/* length of sockaddr{}s */
  int			icmpproto;	/* IPPROTO_XXX value for ICMP */
} ping_t;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	readloop(int);
static void	proc_v4(char *, ssize_t, struct timeval *);
static void	tv_sub(struct timeval *, struct timeval *);
static void	send_v4(void);
static uint16_t	in_cksum(uint16_t *, int);
static void	sig_alarm(int);
#if	0
static ssize_t	send_packet(netfd_t * nfd, const void * msg, size_t len,
			    const struct sockaddr * to, socklen_t tolen,
			    int timeout);
static ssize_t	recv_packet(netfd_t * nfd, void * buf, size_t buflen,
			    struct sockaddr * from, socklen_t * fromlen,
			    int timeout);
#endif

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
/* data that goes with ICMP echo request */
static const int	datalen = DEFDATALEN;
static pid_t		pid;	/* our PID */
static netfd_t *	nfd = NULL;
static sig_atomic_t	recv_timeout = PING_TIMEOUT;

static ping_t	ping_v4 = {proc_v4, send_v4, NULL, NULL, 0, IPPROTO_ICMP};
static ping_t *	ping;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  const char *		host;
  struct addrinfo *	ai;

  if (argc != 2) {
    fprintf(stderr, "usage: %s <hostname>\n", argv[0]);
    exit(1);
  }
  host = argv[argc - 1];

  /* initialize thread engine */
  if (netfd_init(0) == -1) {
    perror("netfd_init");
    exit(1);
  }

  if (netfd_host_serv(host, NULL, PF_INET, SOCK_RAW, IPPROTO_ICMP, &ai) == 0) {
    fprintf(stderr, "PING %s (%s): %d data bytes\n",
	    ai->ai_canonname ? ai->ai_canonname : host,
	    netfd_sock_ntop(ai->ai_addr, false, NULL, 0), datalen);

    /* initialize according to protocol */
    if (ai->ai_family == AF_INET) {
      ping = &ping_v4;
    }
    else {
      fprintf(stderr, "Error: unknown address family %d\n", ai->ai_family);
      exit(1);
    }

    ping->sasend = ai->ai_addr;
    ping->sarecv = calloc(1, ai->ai_addrlen);
    ping->salen  = ai->ai_addrlen;

    readloop(PING_COUNT);

    freeaddrinfo(ai);
  }

  netfd_finish();

  return 0;
}

static void
readloop(max)
	int	max;	/* max count */
{
  int	soc;
  int	size;

  soc = socket(ping->sasend->sa_family, SOCK_RAW, ping->icmpproto);
  if (soc == -1) {
    perror("socket");
    exit(1);
  }
  setuid(getuid());	/* don't need special permissions any more */

  size = IP_MAXPACKET + 128;	/* OK if setsockopt fails */
  (void)setsockopt(soc, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
  nfd = netfd_fdopen(soc);

  pid = getpid() & 0xffff;
  do {
    struct sigaction	act;	/* new */
    struct sigaction	oact;	/* old */

    sigemptyset(&act.sa_mask);
    act.sa_flags   = SA_RESTART;
    act.sa_handler = (void *)sig_alarm;
    sigaction(SIGALRM, &act, &oact);
  } while (0);
//  signal(SIGALRM, sig_alarm);
  do {
    struct timeval	interval = { PING_INTERVAL, 0L };
    struct timeval	value	 = { PING_INTERVAL, 0L };
    struct itimerval	itv	 = { interval, value };

    setitimer(ITIMER_REAL, &itv, NULL);
  } while (0);

  for (; max > 0; max--) {
    socklen_t		len;
    uint8_t		recvbuf[MINICMPLEN + MAXPAYLOAD];
    ssize_t		n;
    struct timeval	tval;
    
    len = ping->salen;
#if	0
    n = netfd_recvfrom(nfd, recvbuf, sizeof(recvbuf), 0, ping->sarecv, &len);
#else
//    n = recv_packet(nfd, recvbuf, sizeof(recvbuf), ping->sarecv, &len, PING_TIMEOUT);
//    n = recv_packet(nfd, recvbuf, sizeof(recvbuf), ping->sarecv, &len, NETFD_NO_TIMEOUT);
    n = netfd_recvfrom(nfd, recvbuf, sizeof(recvbuf), 0, ping->sarecv, &len, recv_timeout);
#endif
    switch (n) {
      case -1:
	if (errno != EINTR) {
  	  perror("netfd_revfrom");
  	  exit(1);
   	}
      case 0:
	if (errno == ETIMEDOUT) {
	  fprintf(stderr, "DEBUG[recv] timedout\n");
   	}
	continue;
      default:
	break;
    }
    gettimeofday(&tval, NULL);
    (*ping->fproc)(recvbuf, n, &tval);
  }
  netfd_close(nfd);
}

static void
proc_v4(ptr, len, tvrecv)
	char *			ptr;
	ssize_t			len;
	struct timeval *	tvrecv;
{
  const struct ip *	ip;
  int			hlen1;
  struct icmp *		icmp;
  int			icmplen;

  ip	= (struct ip *)ptr;	/* start of IP header */
  hlen1	= ip->ip_hl << 2;	/* length of IP header */

  icmp	= (struct icmp *)(ptr + hlen1);	/* start of ICMP header */
  icmplen = len - hlen1;
  if (icmplen < 8) {
    fprintf(stderr, "icmplen (%d) < 8\n", icmplen);
    exit(1);
  }

  if (icmp->icmp_type == ICMP_ECHOREPLY) {
    struct timeval *	tvsend;
    double		rtt;

    if (icmp->icmp_id != pid) {
      return;	/* not a response to our ECHO_REQUEST */
    }
    if (icmplen < 16) {
      fprintf(stderr, "icmplen (%d) < 16\n", icmplen);
      exit(1);
    }

    tvsend = (struct timeval *)icmp->icmp_data;
    tv_sub(tvrecv, tvsend);
    rtt	= tvrecv->tv_sec * 1000.0 + tvrecv->tv_usec / 1000.0;

    fprintf(stdout, "%d bytes from %s: seq=%u ttl=%d rtt=%.3f ms\n",
	    icmplen, netfd_sock_ntop(ping->sarecv, false, NULL, 0),
	    icmp->icmp_seq, ip->ip_ttl, rtt);
  }
}

static void
tv_sub(out, in)
	register struct timeval *	out;
	register struct timeval *	in;
{
  out->tv_usec -= in->tv_usec;
  if (out->tv_usec < 0) {
    --out->tv_sec;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}

static void
send_v4(void)
{
  if (nfd) {
    static int		nsent = 0;	/* add 1 for each sendto() */
    uint8_t		sendbuf[MINICMPLEN + MAXPAYLOAD];
    int			len;
    struct icmp *	icmp;

    icmp = (struct icmp *)sendbuf;
    icmp->icmp_type	= ICMP_ECHO;
    icmp->icmp_code	= 0;
    icmp->icmp_cksum	= 0;
    icmp->icmp_id	= pid;
    icmp->icmp_seq	= nsent++;
    gettimeofday((struct timeval *)icmp->icmp_data, NULL);

    len = MINICMPLEN + datalen;	/* checksum ICMP header and data */
    icmp->icmp_cksum = in_cksum((uint16_t *)icmp, len);

#if	0
    (void)netfd_sendto(nfd, sendbuf, (size_t)len, 0, ping->sasend, ping->salen);
#else
//    (void)send_packet(nfd, sendbuf, (size_t)len, ping->sasend, ping->salen, NETFD_NO_TIMEOUT);
    (void)netfd_sendto(nfd, sendbuf, (size_t)len,0,  ping->sasend, ping->salen, NETFD_NO_TIMEOUT);
    recv_timeout = PING_TIMEOUT;
#endif
  }
}

static uint16_t
in_cksum(addr, len)
	uint16_t *	addr;
	int		len;
{
  int		nleft	= len;
  int		sum	= 0;
  uint16_t *	w	= addr;
  uint16_t	answer	= 0;

  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum),
   * we add sequential 16 bit words to it, and at the end,
   * fold back all the carry bits from the top 16 bits into the lower 16 bits.
   */
  while (nleft > 1) {
    sum	  += *w++;
    nleft -= 2;
  }

  /* 4mop up an odd byte, if neccessary */
  if (nleft == 1) {
#if	1
    *(uint8_t *)(&answer) = *(uint8_t *)w;
    sum += answer;
#else
    union {
      uint16_t	us;
      uint8_t	uc[2];
    } last;
    last.uc[0] = *(uint8_t *)w;
    last.uc[1] = 0;
    sum += last.us;
#endif
  }

  sum	 = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum	+= (sum >> 16);			/* add carry */
  answer = ~sum;			/* truncate to 16 bits */

  return answer;
}

static void
sig_alarm(signo)
	int	signo;
{
  (*ping->fsend)();
}

#if	0
static ssize_t
send_packet(nfd, msg, len, to, tolen, timeout)
	register netfd_t *	nfd;
	const void *		msg;
	size_t			len;
	const struct sockaddr *	to;
	socklen_t		tolen;
	int			timeout;
{
  pollfd_t	pfds;

  pfds.nfd    = nfd;
  pfds.events = NETFD_POLLOUT;
  switch (netfd_poll(&pfds, 1, timeout)) {
    case -1: break;
    case 0:  return 0;
    default:
      if (pfds.revents & NETFD_POLLOUT) {
	ssize_t	n;

	n = netfd_sendto(nfd, msg, len, 0, to, tolen);

	return n;
      }
      break;
  }

  return -1;
}

static ssize_t
recv_packet(nfd, buf, buflen, from, fromlen, timeout)
	register netfd_t *	nfd;
	void *			buf;
	size_t			buflen;
	struct sockaddr *	from;
	socklen_t *		fromlen;
	int			timeout;
{
  pollfd_t	pfds;
  int		rv;

  pfds.nfd    = nfd;
  pfds.events = NETFD_POLLIN;
again:
  switch ((rv = netfd_poll(&pfds, 1, timeout))) {
    case -1:
      if (errno == EINTR) {
	goto again;
      }
      break;
    case 0:
      return 0;
    default:
      if (pfds.revents & NETFD_POLLIN) {
	ssize_t	n;

	n = netfd_recvfrom(nfd, buf, buflen, 0, from, fromlen);

	return n;
      }
      break;
  }

  return -1;
}
#endif
