/*****************************************************************************
 *
 * FILE:	netfd_core.c
 * DESCRIPTION:	Core API for Easy Network I/O Library
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Tue, Mar 18 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd_core.c,v 1.2 2008/03/18 14:39:39 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif	/* HAVE_SYS_STAT_H */
#include <sys/time.h>
#include <sys/event.h>
#include <sys/select.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"
#include "netfd-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/* default value */
#define	DEF_MAXFDS		NETFD_DEFAULT_MODEL
#define	DEF_BLOCKMODE		(1)	/* 0: block, 1: nonblock */

typedef struct _netfd_kq_t {
  int			fd;	/* fro kqueu(2) */
  struct kevent *	events;
  int			nevents;
} netfd_kq_t;
/* netfd virtual processor */
typedef struct _netfd_vp_t {
  int		maxfds;
  int		curfds;
  netfd_t *	freeobj;
  netfd_kq_t	kq;
  netfd_kq_t	poll;
} netfd_vp_t;

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/
static netfd_t *	netfd_new(int osfd, int nonblock);
static void		netfd_free(netfd_t * nfd);
static ssize_t		one_read(netfd_t * nfd, char * ptr,
				 netfd_time_t timeout);
static int		readbuf(netfd_t * nfd, netfd_time_t timeout);

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/
static netfd_vp_t *	nvp = NULL;
static sig_atomic_t	dispatched = 0;	/* flag for netfd_dispatch() */ 

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
int
netfd_init(maxfds)
	int	maxfds;
{
  if (nvp == NULL) {
    /* create netfd_vp_t object */
    nvp = (netfd_vp_t *)malloc(sizeof(netfd_vp_t));
    if (nvp != NULL) {
      /* regulation */
      if (maxfds <= 0) {
	maxfds = DEF_MAXFDS;
      }
      nvp->maxfds  = maxfds;
      nvp->curfds  = 0;
      nvp->freeobj = NULL;

      /* prepare kevent(2) */
      nvp->kq.nevents = 0;
      nvp->kq.events  = (struct kevent *)calloc((size_t)maxfds,
						sizeof(struct kevent));
      if (nvp->kq.events == NULL) {
	goto error1;
      }
      nvp->kq.fd = kqueue();
      if (nvp->kq.fd == -1) {
	goto error2;
      }
      /* poll */
      nvp->poll.nevents = 0;
      nvp->poll.events  = (struct kevent *)calloc((size_t)maxfds,
						  sizeof(struct kevent));
      if (nvp->poll.events == NULL) {
	goto error3;
      }
      nvp->poll.fd = kqueue();
      if (nvp->poll.fd == -1) {
	goto error4;
      }
      return NETFD_INIT_OK;

error4:
      free(nvp->poll.events);
error3:
      close(nvp->kq.fd);
error2:
      free(nvp->kq.events);
error1:
      free(nvp);
      nvp = NULL;
    }
  }
  else {
    return NETFD_INIT_ALREADY;	/* already initialized */
  }

  return NETFD_INIT_ERROR;
}

void
netfd_finish(void)
{
  if (nvp != NULL) {
    my_close(nvp->poll.fd);
    free(nvp->poll.events);
    my_close(nvp->kq.fd);
    free(nvp->kq.events);
    free(nvp);
    nvp = NULL;
  }
}

int
my_close(fd)
	int	fd;
{
  int	status;
  int	saved_errno = errno;
  while (((status = close(fd)) < 0) && (errno == EINTR)) { continue; }
  errno = saved_errno;
  return status;
}

/*****************************************************************************/

netfd_t *
netfd_open(path, flags, mode)
	const char *	path;
	int		flags;
	mode_t		mode;
{
  int	fd;
  netfd_t *	newfd;

  while ((fd = open(path, flags, mode)) < 0) {
    if (errno != EINTR) {
      return NULL;
    }
  }

  newfd = netfd_new(fd, DEF_BLOCKMODE);
  if (newfd == NULL) {
    my_close(fd);
  }

  return newfd;
}

netfd_t *
netfd_fdopen(osfd)
	int	osfd;
{
  return (osfd > 0) ? netfd_new(osfd, DEF_BLOCKMODE) : NULL;
}

static netfd_t *
netfd_new(osfd, nonblock)
	int	osfd;
	int	nonblock;
{
  netfd_t *	nfd;

  if (nvp->freeobj != NULL) {
    nfd		 = nvp->freeobj;
    nvp->freeobj = nvp->freeobj->next;
  }
  else if (nvp->curfds > nvp->maxfds) {
    errno = EMFILE;
    return NULL;
  }
  else {
    nfd = (netfd_t *)malloc(sizeof(netfd_t));
    if (nfd == NULL) {
      errno = ENOMEM;
      return NULL;
    }
    nvp->curfds++;
  }

  /* initialize */
  nfd->osfd	= osfd;
  nfd->rb.total = 0;
  nfd->rb.bufp	= nfd->rb.buf;
  memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
  nfd->path	= NULL;
  nfd->addrlen	= 0;
  nfd->saddr	= NULL;
  nfd->flags	= 0;
  nfd->next	= NULL;
#if	DEBUG
  nfd->count	= 0;
#endif	/* DEBUG */

  if (nonblock) {
    int	flags;

    if ((flags = fcntl(osfd, F_GETFL, 0)) == -1 ||
	fcntl(osfd, F_SETFL, flags|O_NONBLOCK) == -1) {
      netfd_free(nfd);
      nfd = NULL;
    }
  }

  return nfd;
}

static void
netfd_free(nfd)
	netfd_t *	nfd;
{
  if (nfd != NULL) {
#if	DEBUG
    fprintf(stderr, "DEBUG[netfd_free] count=%u", nfd->count);
    if (nfd->flags & NETFD_FL_LISTENER) {
      fputs("[*]", stderr);
    }
    fputc('\n', stderr);
    nfd->count	     = 0;
#endif	/* DEBUG */
    if (nfd->path != NULL) {
      unlink(nfd->path);
      free(nfd->path);
    }
    if (nfd->saddr != NULL) {
      free(nfd->saddr);
    }
    nfd->next	  = nvp->freeobj;
    nvp->freeobj  = nfd;
  }
}

int
netfd_close(nfd)
	netfd_t *	nfd;
{
  int	status = -1;

  if (nfd != NULL) {
    status = my_close(nfd->osfd);
    netfd_free(nfd);
  }
  else {
    errno = EINVAL;
  }

  return status;
}

/*****************************************************************************/

ssize_t
netfd_write(nfd, buf, nbytes, timeout)
	netfd_t *	nfd;
	const void *	buf;
	size_t		nbytes;
	netfd_time_t	timeout;	/* milliseconds */
{
#if	1
  size_t	rest = nbytes;

  while (rest > 0) {
    ssize_t	n;

    n = write(nfd->osfd, buf, rest);
    if (n == -1) {
      pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };

      if (errno == EINTR)			{ continue; }
      if (!NETFD_IO_NOT_READY)			{ return -1; }
      if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
    }
    else {
      rest -= n;
      if (rest == 0) { break; }
      buf = (const void *)((const char *)buf + n);
    }
  }

  return (ssize_t)(nbytes - rest);
#else
  size_t	rest = nbytes;

  while (rest > 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLOUT, 0 };
    ssize_t	n;

    if (netfd_poll(&pfds, 1, timeout) <= 0) { return -1; }

    n = write(nfd->osfd, buf, rest);
    if (n == -1) {
      if (errno == EINTR)	{ continue; }
      if (!NETFD_IO_NOT_READY)	{ return -1; }
    }
    else {
      rest -= n;
      if (rest == 0) { break; }
      buf = (const void *)((const char *)buf + n);
    }
  }

  return (ssize_t)(nbytes - rest);
#endif
}

ssize_t
netfd_read(nfd, buf, nbytes, timeout)
	netfd_t *	nfd;
	void *		buf;
	size_t		nbytes;
	netfd_time_t	timeout;	/* milliseconds */
{
  register ssize_t	n   = 0;
  register char *	ptr = (char *)buf;

  do {
    if (readbuf(nfd, timeout) < 0) { return -1; }
    if (nfd->rb.total == 0)	   { break; }
    *ptr++ = *nfd->rb.bufp++;
    nfd->rb.total--;
  } while ((size_t)n++ < nbytes && nfd->rb.total > 0);

  return n;
}

ssize_t
netfd_readline(nfd, buf, maxlen, timeout)
	netfd_t *	nfd;
	void *		buf;
	size_t		maxlen;
	netfd_time_t	timeout;	/* milliseconds */
{
  register ssize_t	n;
  register char *	ptr = (char *)buf;

  for (n = 1; (size_t)n < maxlen; n++) {
    char	c;
    int		status;

    status = one_read(nfd, &c, timeout);
    if (status == 1) {
      *ptr++ = c;
      if (c == LF)  { break; }
      if (c == EOL) { return 0; }	/* EOF?, null data read */
    }
    else if (status == 0) {
      if (n == 1) { return 0; }		/* EOF, no data read */
      else	  { break; }		/* EOF, some data was read */
    }
    else {
      return -1;			/* error */
    }
  }
  *ptr = EOL;

  return n;
}

static ssize_t
one_read(nfd, ptr, timeout)
	netfd_t *	nfd;
	char *		ptr;
	netfd_time_t	timeout;	/* milliseconds */
{
  if (readbuf(nfd, timeout) < 0) { return -1; }
  if (nfd->rb.total == 0)	 { return  0; }
  nfd->rb.total--;
  *ptr = *nfd->rb.bufp++;

  return 1;
}

static int
readbuf(nfd, timeout)
	netfd_t *	nfd;
	netfd_time_t	timeout;	/* milliseconds */
{
#if	0
  if (nfd->rb.total <= 0) {
    memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
    while ((nfd->rb.total = read(nfd->osfd,
				 nfd->rb.buf, sizeof(nfd->rb.buf))) < 0) {
      pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

      if (errno == EINTR)			{ continue; }
      if (!NETFD_IO_NOT_READY)			{ return -1; }
      if (netfd_poll(&pfds, 1, timeout) <= 0)	{ return -1; }
    }
    nfd->rb.bufp = nfd->rb.buf;
  }

  return nfd->rb.total;
#else
  if (nfd->rb.total <= 0) {
    pollfd_t	pfds = { nfd, NETFD_POLLIN, 0 };

    if (netfd_poll(&pfds, 1, timeout) <= 0) { return -1; }

    memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
    while ((nfd->rb.total = read(nfd->osfd,
				 nfd->rb.buf, sizeof(nfd->rb.buf))) < 0) {
      if (errno == EINTR)	{ continue; }
      if (!NETFD_IO_NOT_READY)	{ return -1; }
    }
    nfd->rb.bufp = nfd->rb.buf;
  }

  return nfd->rb.total;
#endif
}

/*****************************************************************************/

off_t
netfd_seek(nfd, offset, whence)
	netfd_t *	nfd;
	off_t	offset;
	int	whence;
{
  struct stat	sbuf;

  if (fstat(nfd->osfd, &sbuf) == 0 && S_ISREG(sbuf.st_mode)) {
    off_t	pos;

    pos = lseek(nfd->osfd, offset, whence);
    if (pos != -1) {
      return pos;
    }
    errno = EINVAL;
  }

  return -1;
}


netfd_t *
netfd_mkstemp(template)
	char *	template;
{
  int	fd;

  fd = mkstemp(template);
  return (fd == -1) ? NULL : netfd_new(fd, DEF_BLOCKMODE);
}

/*****************************************************************************/

/*
 * XXX:
 *	̵ LAN ͥåȥĶǤϡbpf ǥХФ
 *	kqueue(2)/kevent(2) ̤ưʤä
 *	餯PLCPץꥢ֥ PLCPإåˤƶȿ¬롣
 *	̵ LAN ĶѤʤΤǤСʲ #if  0 ˤɤ
 */
#if	1
int
netfd_poll(pfds, pfdnum, timeout)
	pollfd_t	pfds[];
	size_t		pfdnum;
	netfd_time_t	timeout;	/* milliseconds */
{
  int	n = -1;	/* return value */

  if (pfdnum > 0) {
    int			fds = 0;
    fd_set		rfds;
    fd_set		wfds;
    struct timeval	to;
    struct timeval *	tsp;
    struct timeval	tv_st;
    register size_t	i;

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    for (i = 0; i < pfdnum; i++) {
      if (pfds[i].events & NETFD_POLLIN)  { FD_SET(pfds[i].nfd->osfd, &rfds); }
      if (pfds[i].events & NETFD_POLLOUT) { FD_SET(pfds[i].nfd->osfd, &wfds); }
      pfds[i].revents = 0;
      if (pfds[i].nfd->osfd > fds) { fds = pfds[i].nfd->osfd; }
    }

    if (timeout == 0) {
      to.tv_sec	 = 0L;
      to.tv_usec = 0L;
      tsp	 = &to;
    }
    else if (timeout < 0) {
      tsp = NULL;
    }
    else {
      to.tv_sec	 = NETFD_MSEC2SEC(timeout);
      to.tv_usec = NETFD_MSEC2USEC(timeout);
      tsp	 = &to;
    }

    gettimeofday(&tv_st, (struct timezone *)NULL);
    while (((n = select(fds + 1, &rfds, &wfds, NULL, tsp)) < 0) &&
    	   (errno == EINTR)) {
      /* re-calculate timeout */
      if (tsp != NULL) {
	struct timeval	tv_ed;
	struct timeval	tv_sub;

	gettimeofday(&tv_ed, (struct timezone *)NULL);
	timersub(&tv_ed, &tv_st, &tv_sub);
	tv_st = tv_ed;
	to.tv_sec  -= tv_sub.tv_sec;
	to.tv_usec -= tv_sub.tv_usec;
	if (to.tv_usec < 0) {
	  to.tv_sec--;
	  to.tv_usec += 1000000;
	}
	if (to.tv_sec < 0) {
	  to.tv_sec = 0;
	}
      }
      errno = 0;
    }
    switch (n) {
      case 0:	/* timeout */
	errno = ETIMEDOUT;
	break;
      case -1:	/* error */
	if (errno == EBADF) {
	  for (i = 0; i < pfdnum; i++) {
	    pfds[i].revents = 0;
	    if (fcntl(pfds[i].nfd->osfd, F_GETFL, 0) == -1) {
	      pfds[i].revents = NETFD_POLLNVAL;
	    }
	  }
	}
	break;
      default:
	for (i = 0; i < (size_t)n; i++) {
	  if (FD_ISSET(pfds[i].nfd->osfd, &rfds)) {
	    pfds[i].revents |= NETFD_POLLIN;
	  }
	  if (FD_ISSET(pfds[i].nfd->osfd, &wfds)) {
	    pfds[i].revents |= NETFD_POLLOUT;
	  }
	}
	break;
    }
  }

  return n;
}
#else
int
netfd_poll(pfds, pfdnum, timeout)
	pollfd_t	pfds[];
	size_t		pfdnum;
	netfd_time_t	timeout;	/* milliseconds */
{
  int	n = -1;	/* return value */

  if (pfdnum > 0) {
    struct timespec	to;
    struct timespec *	tsp;
    struct timeval	tv_st;
    register size_t	i;

    /* set events */
    nvp->poll.nevents = 0;
    for (i = 0; i < pfdnum; i++) {
      if (pfds[i].events & NETFD_POLLIN) {
	EV_SET(&nvp->poll.events[nvp->poll.nevents++], pfds[i].nfd->osfd,
		 EVFILT_READ, EV_ADD | EV_ONESHOT, 0, 0, pfds[i].nfd);
      }
      if (pfds[i].events & NETFD_POLLOUT) {
	EV_SET(&nvp->poll.events[nvp->poll.nevents++], pfds[i].nfd->osfd,
		 EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, pfds[i].nfd);
      }
      pfds[i].revents = 0;
    }

    /* calulate timeout */
    if (timeout == 0) {
	to.tv_sec	= 0L;
	to.tv_nsec	= 0L;
	tsp		= &to;
    }
    else if (timeout < 0) {
	tsp = NULL;
    }
    else {
	to.tv_sec	= NETFD_MSEC2SEC(timeout);
	to.tv_nsec	= NETFD_MSEC2NSEC(timeout);
	tsp		= &to;
    }

    gettimeofday(&tv_st, (struct timezone *)NULL);
    while (((n = kevent(nvp->poll.fd, nvp->poll.events, nvp->poll.nevents,
			nvp->poll.events, (int)pfdnum, tsp)) < 0) &&
    	    (errno == EINTR)) {
      /* re-calculate timeout */
      if (tsp != NULL) {
	struct timeval	tv_ed;
	struct timeval	tv_sub;

	gettimeofday(&tv_ed, (struct timezone *)NULL);
	timersub(&tv_ed, &tv_st, &tv_sub);
	tv_st = tv_ed;
	to.tv_sec  -= tv_sub.tv_sec;
	to.tv_nsec -= (tv_sub.tv_usec * 1000);
	if (to.tv_nsec < 0) {
	  to.tv_sec--;
	  to.tv_nsec += 1000000000;
	}
	if (to.tv_sec < 0) {
	  to.tv_sec = 0;
	}
      }
      errno = 0;
    }
    switch (n) {
      case 0:	/* timeout */
	errno = ETIMEDOUT;
	break;
      case -1:	/* error */
	if (errno == EBADF) {
	  for (i = 0; i < pfdnum; i++) {
	    pfds[i].revents = 0;
	    if (fcntl(pfds[i].nfd->osfd, F_GETFL, 0) == -1) {
	      pfds[i].revents = NETFD_POLLNVAL;
	    }
	  }
	}
	break;
      default:
	for (i = 0; i < (size_t)n; i++) {
	  if ((pfds[i].events & NETFD_POLLIN) &&
	      nvp->poll.events[i].ident  == (uintptr_t)pfds[i].nfd->osfd &&
	      nvp->poll.events[i].filter == EVFILT_READ) {
	    pfds[i].revents  |= NETFD_POLLIN;
	  }
	  if ((pfds[i].events & NETFD_POLLOUT) &&
	      nvp->poll.events[i].ident  == (uintptr_t)pfds[i].nfd->osfd &&
	      nvp->poll.events[i].filter == EVFILT_WRITE) {
	    pfds[i].revents  |= NETFD_POLLOUT;
	  }
	}
	break;
    }
  }

  return n;
}
#endif

/*****************************************************************************/

int
netfd_dispatch(callback, writeshot, timeout)
	netfd_callback_t	callback;
	const bool		writeshot;
				/* true: enable EV_WRITE|EV_ONESHOT */
	netfd_time_t		timeout;	/* milliseconds */
{
  register int		i;
  int			n;
  struct timespec	to;
  struct timespec *	tsp;

  if (dispatched) { return -1; }

  if (timeout == 0) {
    to.tv_sec	= 0L;
    to.tv_nsec	= 0L;
    tsp		= &to;
  }
  else if (timeout < 0) {
    tsp = NULL;
  }
  else {
    to.tv_sec	= NETFD_MSEC2SEC(timeout);
    to.tv_nsec	= NETFD_MSEC2NSEC(timeout);
    tsp		= &to;
  }

  dispatched = 1;
  do {
    n = kevent(nvp->kq.fd, nvp->kq.events, nvp->kq.nevents,
	       nvp->kq.events, nvp->maxfds, tsp);
    switch (n) {
      case -1: break;
      case 0:  errno = ETIMEDOUT; break;
      default:
	nvp->kq.nevents = 0;
	for (i = 0; i < n; i++) {
	  netfd_t *	nfd = (netfd_t *)nvp->kq.events[i].udata;

	  if (nfd->flags & NETFD_FL_LISTENER) {
	    struct sockaddr 	cliaddr;
	    socklen_t		clilen;
	    int			sd;

	    sd = accept(nfd->osfd, &cliaddr, &clilen);
	    if (sd != -1) {
	      netfd_t *	newfd;
#ifdef	SO_NOSIGPIPE
	      const int	on = 1;
	      /* Do not generate SIGPIPE */
	      (void)setsockopt(sd, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on));
#endif	/* SO_NOSIGPIPE */

	      newfd = netfd_fdopen(sd);
	      if (newfd != NULL) {
		newfd->saddr = (struct sockaddr *)malloc(clilen);
		if (newfd->saddr != NULL) {
		  memcpy(newfd->saddr, &cliaddr, clilen);
		  newfd->addrlen = clilen;
		}
		else {
		  netfd_close(newfd);
		  break;	/* exit switch */
		}
		EV_SET(&nvp->kq.events[nvp->kq.nevents++], newfd->osfd,
		       EVFILT_READ, EV_ADD, 0, 0, newfd);
		/*
		 * Generally, server writes data after it reads data.
		 * This writeshot flag enable server to write data immediately
		 * without reading any data.
		 */
		if (writeshot) {
		  EV_SET(&nvp->kq.events[nvp->kq.nevents++], newfd->osfd,
			 EVFILT_WRITE, EV_ADD | EV_ONESHOT, 0, 0, newfd);
		}
	      }
	      else {
		my_close(sd);
	      }
	    }
	  }
	  else if (nvp->kq.events[i].flags == EV_EOF) {
	    netfd_close(nfd);
	  }
	  else {
#if	DEBUG
	    nfd->count++;
#endif	/* DEBUG */
	    (*callback)(nfd);
	  }
	}
	break;
    }
  } while (n > 0);
  dispatched = 0;

  return n;	/* -1: error, 0: timeout */
}

void
netfd_event_set(nfd, filter, flags)
	netfd_t *	nfd;
	int		filter;
	int		flags;
{
  EV_SET(&nvp->kq.events[nvp->kq.nevents++], nfd->osfd, filter, flags,
	 0, 0, nfd);
}
