/*****************************************************************************
 *
 * FILE:	netfd_icmp.c
 * DESCRIPTION:	ICMP Echo/Reply API for Easy Network I/O Library
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Tue, Mar 18 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd_icmp.c,v 1.1 2008/03/18 13:31:19 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include "netfd.h"
#include "netfd-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	PHDR_LEN	((int)sizeof(struct timeval))
#define	DEFDATALEN	(64 - PHDR_LEN)	/* default data length */
#define	MAXIPLEN	(sizeof(struct ip) + MAX_IPOPTLEN)
#define	MINICMPLEN	ICMP_MINLEN
#define	MAXPAYLOAD	(IP_MAXPACKET - MAXIPLEN - MINICMPLEN)

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/
static void	tv_sub(struct timeval * out, struct timeval * in);
static int	proc_icmp(uint8_t * buf, ssize_t len, uint16_t check_id,
			  struct sockaddr * from, struct timeval * tvrecv);

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
netfd_t *
netfd_icmp_open(void)
{
  int	sd;

  sd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (sd != -1) {
    static const int	size = IP_MAXPACKET + 128;
    netfd_t *		nfd;

    /* OK even if setsockopt fails */
    (void)setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
    if (getuid() == 0) {
      (void)setsockopt(sd, SOL_SOCKET, SO_SNDBUF, &size, sizeof(size));
    }
    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

ssize_t
netfd_icmp_send(nfd, to, tolen, send_id, seqno, timeout)
	netfd_t *		nfd;
	const struct sockaddr *	to;
	socklen_t		tolen;
	uint16_t		send_id;	/* icmp_id */
	uint32_t		seqno;		/* icmp_seq */
	netfd_time_t		timeout;	/* milliseconds */
{
  struct icmp *	icmp;
  uint8_t	sendbuf[MINICMPLEN + MAXPAYLOAD];
  size_t	len;

  /* make ICMP packet */
  icmp = (struct icmp *)sendbuf;
  icmp->icmp_type	= ICMP_ECHO;
  icmp->icmp_code	= 0;
  icmp->icmp_cksum	= 0;
  icmp->icmp_id		= send_id;
  icmp->icmp_seq	= seqno;
  gettimeofday((struct timeval *)icmp->icmp_data, NULL);

  len = MINICMPLEN + DEFDATALEN;	/* checksum ICMP header and data */
  icmp->icmp_cksum = my_in_cksum((uint16_t *)icmp, len);

  return netfd_sendto(nfd, sendbuf, (size_t)len, 0, to, tolen, timeout);
}

uint16_t
my_in_cksum(addr, len)
	uint16_t *	addr;
	size_t		len;
{
  int		nleft	= len;
  int		sum	= 0;
  uint16_t *	w	= addr;
  uint16_t	answer	= 0;

  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum),
   * we add sequential 16 bit words to it, and at the end,
   * fold back all the carry bits from the top 16 bits into the lower 16 bits.
   */
  while (nleft > 1) {
    sum	  += *w++;
    nleft -= 2;
  }

  /* 4mop up an odd byte, if neccessary */
  if (nleft == 1) {
    *(uint8_t *)(&answer) = *(uint8_t *)w;
    sum += answer;
  }

  sum	 = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum	+= (sum >> 16);				/* add carry */
  answer = ~sum;				/* truncate to 16 bits */

  return answer;
}

ssize_t
netfd_icmp_recv(nfd, from, fromlen, check_id, timeout, callback)
	netfd_t *		nfd;
	struct sockaddr *	from;
	socklen_t *		fromlen;
	uint16_t		check_id;	/* icmp id */
	netfd_time_t		timeout;	/* milliseconds */
	netfd_icmp_callback_t	callback;
{
  uint8_t	recvbuf[MINICMPLEN + MAXPAYLOAD];
  ssize_t	n;

  n = netfd_recvfrom(nfd, recvbuf, sizeof(recvbuf), 0, from, fromlen, timeout);
  if (n > 0) {
    struct timeval	now;

    gettimeofday(&now, (struct timezone *)NULL);
    if (callback == NULL) {
      callback = proc_icmp;
    }
    n = (*callback)(recvbuf, n, check_id, from, &now);
  }
#if	DEBUG0
  fprintf(stderr, "DEBUG[netfd_icmp_recv] n=%d errno=%d\n", n, errno);
#endif	/* DEBUG */

  return n;
}

static int
proc_icmp(buf, len, check_id, from, tvrecv)
	uint8_t *		buf;
	ssize_t			len;
	uint16_t		check_id;	/* checked icmp id */
	struct sockaddr *	from;
	struct timeval *	tvrecv;
{
  const struct ip *	ip;
  int			hlen;
  const struct icmp *	icmp;
  int			icmplen;

  ip	= (const struct ip *)buf;	/* start of IP header */
  hlen	= ip->ip_hl << 2;		/* length of IP header */

  icmp	= (const struct icmp *)(buf + hlen);	/* start of ICMP header */
  icmplen = len - hlen;
  if (icmplen < 8) {
    fprintf(stderr, "icmplen (%d) < 8\n", icmplen);
    return -1;
  }

  if (icmp->icmp_type == ICMP_ECHOREPLY) {
    struct timeval *	tvsend;
    double		rtt;

    if (icmp->icmp_id != check_id) {
      return -1;	/* not a response to our ECHO_REQUEST */
    }
    if (icmplen < 16) {
      fprintf(stderr, "icmplen (%d) < 16\n", icmplen);
      return -1;
    }

    tvsend = (struct timeval *)(icmp->icmp_data);
    tv_sub(tvrecv, tvsend);
    rtt	= tvrecv->tv_sec * 1000.0 + tvrecv->tv_usec / 1000.0;

    fprintf(stdout, "%d bytes from %s: seq=%u ttl=%d rtt=%.3f ms\n",
	    icmplen, netfd_sock_ntop(from, false, NULL, 0),
	    icmp->icmp_seq, ip->ip_ttl, rtt);
    return 0;
  }

  return -1;
}

static void
tv_sub(out, in)
	register struct timeval *	out;
	register struct timeval *	in;
{
  out->tv_usec -= in->tv_usec;
  if (out->tv_usec < 0) {
    --out->tv_sec;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}
