/*****************************************************************************
 *
 * FILE:	netfd_link.c
 * DESCRIPTION:	Link Layer Access API for Easy Network I/O Library
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Tue, Mar 18 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd_link.c,v 1.1 2008/03/18 13:31:19 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif	/* HAVE_SYS_IOCTL_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <arpa/inet.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/bpf.h>
#include <net/ethernet.h>
#include <netinet/if_ether.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"
#include "netfd-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#ifdef	ETHER_HDR_LEN
#define	SIZE_ETHERHDR	ETHER_HDR_LEN
#else
#define	SIZE_ETHERHDR	(sizeof(struct ether_header))
#endif	/* ETHER_HDR_LEN */

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/
static int	get_ifconf(int sd, struct ifconf * ifc, char ** buf);

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
netfd_t *
netfd_open_bpf(netif, flags, capsize, promiscuous, to_ms, pktfilter)
	const char *		netif;		/* network interface */
	uint16_t		flags;		/* bpf flags */
	size_t			capsize;	/* buffer size to capture */
	bool			promiscuous;	/* promiscus: true|false */
	netfd_time_t		to_ms;		/* timeout (milliseconds) */
	struct bpf_program *	pktfilter;	/* packet filter */
{
  if (netif != NULL) {
    int	fd = -1;
    int	n  = 0;

    do {
      char	s[sizeof("/dev/bpfXXX")];

      (void)snprintf(s, sizeof(s), "/dev/bpf%d", n++);
      fd = open(s, O_RDWR);
    } while (fd == -1 && (errno == EBUSY || errno == EPERM));

    if (fd != -1) {
      struct ifreq	ifreq;

      /* set buffer size to capture */
      if (capsize > 0) {
	size_t	bufsize;

	for (bufsize = capsize; bufsize != 0; bufsize >>= 1) {
	  if (ioctl(fd, BIOCSBLEN, &bufsize) == 0) { break; }
	}
	if (bufsize == 0) { goto error; }
      }
      /* set network device */
      memset(&ifreq, 0, sizeof(ifreq));
      strlcpy(ifreq.ifr_name, netif, sizeof(ifreq.ifr_name));
      if ( ioctl(fd, BIOCSETIF, &ifreq) == -1) { goto error; }
      /* set packet filter */
      if (pktfilter != NULL) {
	if (ioctl(fd, BIOCSETF, pktfilter) == -1) { goto error; }
      }
      /* set promiscus mode */
      if (promiscuous) {
       	if (ioctl(fd, BIOCPROMISC, NULL) == -1) { goto error; }
      }
      /* set timeout */
      if (to_ms > 0) {
	struct timeval	timeout;

	timeout.tv_sec  = NETFD_MSEC2SEC(to_ms);
	timeout.tv_usec = NETFD_MSEC2USEC(to_ms);
	if (ioctl(fd, BIOCSRTIMEOUT, &timeout) == -1) { goto error; }
      }
#ifdef	BIOCIMMEDIATE
      /* no block in kernel buffer */
      else if (flags & NETFD_BPF_IMMEDIATE) {
	static const unsigned int	on = 1;
	if (ioctl(fd, BIOCIMMEDIATE, &on) == -1) { goto error; }
      }
#endif	/* BIOCIMMEDIATE */
      /* build ether frame by yourself */
      if (flags & NETFD_BPF_HDRCMPLT) {
	static const unsigned int	on = 1;
	if (ioctl(fd, BIOCSHDRCMPLT, &on) == -1) { goto error; }
      }
      /* see only received packets */
      if (flags & NETFD_BPF_NOSEESENT) {
	static const unsigned int	off = 0;
	if (ioctl(fd, BIOCSSEESENT, &off) == -1) { goto error; }
      }
      if (ioctl(fd, BIOCFLUSH, 0) == 0) {
	netfd_t *	nfd;

	nfd = netfd_fdopen(fd);
	if (nfd != NULL) {
	  nfd->flags |= NETFD_FL_BPF;
	  return nfd;
	}
      }
error:
      my_close(fd);
    }
  }

  return NULL;
}

ssize_t
netfd_build_ether_frame(type, src_host, dst_host, data, datasize, frame)
	uint16_t	type;		/* ether type */
	const char *	src_host;	/* source host address */
	const char *	dst_host;	/* destination host address */
	uint8_t *	data;		/* paylaod */
	size_t		datasize;	/* payload size */
	uint8_t **	frame;		/* returned */
{
  if (frame != NULL) {
    size_t	framesize;

    framesize = SIZE_ETHERHDR + datasize;
    *frame    = (uint8_t *)malloc(framesize);
    if (*frame != NULL) {
      struct ether_header *	ether;
      struct ether_addr		shost;
      struct ether_addr		dhost;

      memset(*frame, 0, framesize);
      ether = (struct ether_header *)(*frame);

      /* build header */
      shost = *(struct ether_addr *)ether_aton(src_host);
      dhost = *(struct ether_addr *)ether_aton(dst_host);
      ether->ether_type = htons(type);
      memcpy(ether->ether_shost, shost.octet, ETHER_ADDR_LEN);
      memcpy(ether->ether_dhost, dhost.octet, ETHER_ADDR_LEN);

      /* copy payload */
      memcpy(*frame + SIZE_ETHERHDR, data, datasize);

      return framesize;
    }
  }

  return -1;
}

/*****************************************************************************/

netfd_ifinfo_t *
netfd_get_ifinfo(void)
{
  netfd_ifinfo_t *	ifihead = NULL;
  int			sd;

  sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
  if (sd != -1) {
    struct ifconf	ifc;
    char *		buf;

    if (get_ifconf(sd, &ifc, &buf) == 0) {
      netfd_ifinfo_t **	ifi_next;
      char		lastname[IFNAMSIZ];
      sa_family_t	lastfamily = AF_UNSPEC;
      char *		ptr;

      ifi_next	  = &ifihead;
      lastname[0] = '\0';

      for (ptr = buf; ptr < buf + ifc.ifc_len; ) {
	struct ifreq *		ifr;
	struct ifreq		ifrcopy;
	int			len;
	netfd_ifinfo_t *	ifi;
	int			flags;
	int			myflags;
	char *			cptr;
	sa_family_t		family;

	ifr  = (struct ifreq *)ptr;
	len  = max(sizeof(struct sockaddr), ifr->ifr_addr.sa_len);
	ptr += sizeof(ifr->ifr_name) + len;	/* for next one in buffer */

	myflags = 0;
	if ( (cptr = strchr(ifr->ifr_name, ':')) != NULL) {
	  *cptr = 0;		/* replace colon will null */
	}
	family = ifr->ifr_addr.sa_family;
	if (strncmp(lastname, ifr->ifr_name, IFNAMSIZ) == 0 &&
	    lastfamily == family) {
	  if (family == AF_INET || family == AF_INET6) {
	    myflags = IFI_ALIAS;
	  }
	}
	if (family != AF_LINK) {
	  memcpy(lastname, ifr->ifr_name, IFNAMSIZ);
	}
	lastfamily = family;

	ifrcopy = *ifr;
	ioctl(sd, SIOCGIFFLAGS, &ifrcopy);
	flags = ifrcopy.ifr_flags;
	if ((flags & IFF_UP) == 0) {
	  continue;	/* ignore if interface not up */
	}

	ifi = calloc(1, sizeof(struct ifi_info));
	if (ifi == NULL) { continue; }
	*ifi_next = ifi;		/* prev points to this new one */
	ifi_next  = &ifi->ifi_next;	/* pointer to next one goes here */

	ifi->ifi_flags	 = flags;	/* IFF_xxx values */
	ifi->ifi_myflags = myflags;	/* IFI_xxx values */
	memcpy(ifi->ifi_name, ifr->ifr_name, IFI_NAME);
	ifi->ifi_name[IFI_NAME-1] = '\0';

	switch (family) {
       	  case AF_INET: {
	      struct sockaddr_in *	sa;

	      sa = (struct sockaddr_in *)&ifr->ifr_addr;
	      if (ifi->ifi_addr == NULL) {
		ifi->ifi_addr = calloc(1, sizeof(struct sockaddr_in));
		if (ifi->ifi_addr != NULL) {
		  memcpy(ifi->ifi_addr, sa, sizeof(struct sockaddr_in));
		}

		if (flags & IFF_BROADCAST) {
		  ioctl(sd, SIOCGIFBRDADDR, &ifrcopy);
		  sa = (struct sockaddr_in *)&ifrcopy.ifr_broadaddr;
		  ifi->ifi_brdaddr = calloc(1, sizeof(struct sockaddr_in));
		  if (ifi->ifi_brdaddr != NULL) {
		    memcpy(ifi->ifi_brdaddr, sa, sizeof(struct sockaddr_in));
		  }
		}

		if (flags & IFF_POINTOPOINT) {
		  ioctl(sd, SIOCGIFDSTADDR, &ifrcopy);
		  sa = (struct sockaddr_in *)&ifrcopy.ifr_dstaddr;
		  ifi->ifi_dstaddr = calloc(1, sizeof(struct sockaddr_in));
		  if (ifi->ifi_dstaddr != NULL) {
		    memcpy(ifi->ifi_dstaddr, sa, sizeof(struct sockaddr_in));
		  }
		}
	      }
	    }
    	    break;
	  case AF_INET6: {
	      struct sockaddr_in6 *	sa;

	      sa = (struct sockaddr_in6 *)&ifr->ifr_addr;
	      if (ifi->ifi_addr == NULL) {
		ifi->ifi_addr = calloc(1, sizeof(struct sockaddr_in6));
		if (ifi->ifi_addr != NULL) {
		  memcpy(ifi->ifi_addr, sa, sizeof(struct sockaddr_in6));
		}
	      }
	    }
	    break;
	  case AF_LINK: {
	      struct sockaddr_dl *	sa;

	      sa = (struct sockaddr_dl *)&ifr->ifr_addr;
	      if (sa->sdl_type == IFT_ETHER && sa->sdl_alen > 0) {
		ifi->ifi_hlen = sa->sdl_alen;
		memset(ifi->ifi_haddr, 0, sizeof(ifi->ifi_haddr));
		memcpy(ifi->ifi_haddr, (uint8_t *)LLADDR(sa), sa->sdl_alen);
	      }
	      if (ifi->ifi_addr == NULL) {
		ifi->ifi_addr = calloc(1, sizeof(struct sockaddr_dl));
		if (ifi->ifi_addr != NULL) {
		  memcpy(ifi->ifi_addr, sa, sizeof(struct sockaddr_dl));
		}
	      }
	    }
	    break;
	  default:
    	    break;
	}
      }
      free(buf);
    }
    my_close(sd);
  }

  return ifihead;	/* pointer to first structure in linked list */
}

void
netfd_free_ifinfo(netfd_ifinfo_t * ifihead)
{
  netfd_ifinfo_t *	ifi;
  netfd_ifinfo_t *	ifi_next;

  for (ifi = ifihead; ifi != NULL; ifi = ifi_next) {
    if (ifi->ifi_addr    != NULL) { free(ifi->ifi_addr); }
    if (ifi->ifi_brdaddr != NULL) { free(ifi->ifi_brdaddr); }
    if (ifi->ifi_dstaddr != NULL) { free(ifi->ifi_dstaddr); }
    ifi_next = ifi->ifi_next;	/* can't fetch ifi_next after free() */
    free(ifi);			/* the ifi_info{} itself */
  }
}

static int
get_ifconf(sd, ifc, buf)
	int		sd;
	struct ifconf *	ifc;	/* returned */
	char **		buf;	/* returned */
{
  size_t	lastlen	= 0;
  size_t	len;
  
  len = 100 * sizeof(struct ifreq);	/* initial buffer size guess */
  while (1) {
    *buf = malloc(len);
    if (*buf == NULL) {
      return -1;
    }
    ifc->ifc_len = len;
    ifc->ifc_buf = *buf;
    if (ioctl(sd, SIOCGIFCONF, ifc) < 0) {
      if (errno != EINVAL || lastlen != 0) {
	return -1;
      }
    }
    else {
      if ((size_t)ifc->ifc_len == lastlen) {
	break;		/* success, len has not changed */
      }
      lastlen = ifc->ifc_len;
    }
    len += 10 * sizeof(struct ifreq);	/* increment */
    free(*buf);
  }

  return 0;
}
