/*****************************************************************************
 *
 * FILE:	netfd_server.c
 * DESCRIPTION:	TCP/UDP Server API for Easy Network I/O Library
 * DATE:	Sat, Jun 10 2006
 * UPDATED:	Mon, Mar 17 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: netfd_server.c,v 1.1 2008/03/18 13:31:19 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <arpa/inet.h>
#include <sys/un.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <errno.h>
#include "netfd.h"
#include "netfd-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/* default value */
#define	DEF_BACKLOG	(8)

/******************************************************************************
 *
 *	Lobal functions declaration
 *
 *****************************************************************************/
static int	netfd_server(const char * hostname, const char * service,
			     int socktype, int protocol, int flags,
			     struct sockaddr ** saddr, socklen_t * addrlen);

/******************************************************************************
 *
 *	Lobal variable definition
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
netfd_t *
netfd_tcp_server(hostname, service, backlog)
	const char *	hostname;
	const char *	service;
	const int	backlog;
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_server(hostname, service, SOCK_STREAM, IPPROTO_TCP, AI_PASSIVE,
		    &saddr, &addrlen);
  if (sd != -1) {
    if (listen(sd, backlog > 0 ? backlog : DEF_BACKLOG) == 0) {
      netfd_t *	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	nfd->addrlen = addrlen;
	nfd->saddr   = saddr;
	nfd->flags  |= NETFD_FL_LISTENER;
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_udp_server(hostname, service)
	const char *	hostname;
	const char *	service;
{
  int			sd;
  struct sockaddr *	saddr;
  socklen_t		addrlen;

  sd = netfd_server(hostname, service, SOCK_DGRAM, IPPROTO_UDP, AI_PASSIVE,
		    &saddr, &addrlen);
  if (sd != -1) {
    netfd_t *	nfd;

    nfd = netfd_fdopen(sd);
    if (nfd != NULL) {
      nfd->addrlen = addrlen;
      nfd->saddr   = saddr;
      return nfd;
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_ust_server(path, backlog)
	const char *	path;
	const int	backlog;
{
  int	sd;

  sd = socket(AF_LOCAL, SOCK_STREAM, 0);
  if (sd != -1) {
    struct sockaddr_un	saddr;

    unlink(path);
    memset((char *)&saddr, 0, sizeof(saddr));
    saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
    strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
    strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
    if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == 0 &&
	listen(sd, backlog > 0 ? backlog : DEF_BACKLOG) == 0) {
      netfd_t *	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

netfd_t *
netfd_udg_server(path)
	const char *	path;
{
  int	sd;

  sd = socket(AF_LOCAL, SOCK_DGRAM, 0);
  if (sd != -1) {
    struct sockaddr_un	saddr;

    unlink(path);
    memset((char *)&saddr, 0, sizeof(saddr));
    saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
    strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
    strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
    if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == 0) {
      netfd_t *	nfd;

      nfd = netfd_fdopen(sd);
      if (nfd != NULL) {
	return nfd;
      }
    }
    my_close(sd);
  }

  return NULL;
}

static int
netfd_server(hostname, service, socktype, protocol, flags, saddr, addrlen)
	const char *		hostname;
	const char *		service;
	int			socktype;
	int			protocol;
	int			flags;
	struct sockaddr **	saddr;		/* returned */
	socklen_t *		addrlen;	/* returned */
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  int			sd = -1;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= socktype;
  hints.ai_protocol	= protocol;
  hints.ai_flags	= flags;

  if (getaddrinfo(hostname, service, &hints, &res) == 0) {
    struct addrinfo *	res_save;

    res_save = res;
    do {
      static const int	on = 1;

      /* create a socket */
      sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
      if (sd == -1) {
	continue;	/*error, try next one */
      }
      /* allow reuse of local address */
      (void)setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

      /* bind a socket */
      if (bind(sd, res->ai_addr, res->ai_addrlen) == 0) {
	*addrlen = res->ai_addrlen;
	*saddr	  = (struct sockaddr *)malloc(res->ai_addrlen);
	if (*saddr != NULL) {
	  memcpy(*saddr, res->ai_addr, res->ai_addrlen);
	}
	break;	/* success */
      }
      my_close(sd);	/* bind error, close and try next one */
      sd = -1;
    } while ((res = res->ai_next) != NULL);
    freeaddrinfo(res_save);
  }

  return sd;
}
